"""
Helper for ViUR-core types and behavior
"""

import logging
import typing as t

from viur.core import current, db
from viur.core.skeleton import SkeletonInstance, skeletonByKind

__all__ = [
    "change_language",
    "get_task_retry_count",
    "without_render_preparation",
    "get_full_skel_from_ref_skel",
    "iter_skel",
]

logger = logging.getLogger(__name__)


def change_language(lang: str) -> None:
    """Change the current language"""
    current.session.get()["lang"] = lang
    current.language.set(lang)


def get_task_retry_count() -> int:
    """Return the number of times the current task is retried as int"""
    try:
        return int(current.request.get().request.headers.get("X-Appengine-Taskretrycount", -1))
    except AttributeError:
        # During warmup current.request is None (at least on local devserver)
        return -1


def without_render_preparation(skel: SkeletonInstance, full_clone: bool = False) -> SkeletonInstance:
    """Return the SkeletonInstance without renderPreparation.

    This method is useful (and unfortunately necessary due to the ViUR design)
    if you call python methods from the jinja template that should work on the
    `SkeletonInstance.accessedValues` and not on the `SkeletonInstance.renderAccessedValues`.

    If the SkeletonInstance does not have renderPreparation, it will be returned as is.
    If renderPreparation is enabled, a new SkeletonInstance is created.
    However, unless `full_clone` is True, the SkeletonInstance will use the
    identical objects as the source skeleton. It just "removes" the
    "renderPreparation mode" and keep it for the source skel enabled.
    """
    if skel.renderPreparation is not None:
        # TODO: ViUR, I DONT WANT TO HAVE RENDERPREPARATION ON MODULE LAYER!!!
        if full_clone:
            skel = skel.clone()
        else:
            src_skel = skel
            # Create a new SkeletonInstance with the same object,
            # but without enabled renderPreparation
            skel = SkeletonInstance(src_skel.skeletonCls, bone_map=src_skel.boneMap)
            skel.accessedValues = src_skel.accessedValues
            skel.dbEntity = src_skel.dbEntity
            skel.errors = src_skel.errors
            skel.is_cloned = src_skel.is_cloned
        assert skel.renderPreparation is None
    return skel


def get_full_skel_from_ref_skel(ref_skel: SkeletonInstance) -> SkeletonInstance:
    kind_name = ref_skel.skeletonCls.__name__.removeprefix("RefSkelFor")
    skel: SkeletonInstance = skeletonByKind(kind_name)()  # noqa
    skel.fromDB(ref_skel["key"])
    return skel


def iter_skel(query: db.Query) -> t.Iterator[SkeletonInstance]:
    """Fetch all entries for this query and yield the skel

    Doesn't use fetch() due to the strange ViUR fetch limit (100).
    Acts as generator to be not memory hungry...
    """
    skel: SkeletonInstance = query.srcSkel
    for entry in query.iter():
        skel = SkeletonInstance(skel.skeletonCls, clonedBoneMap=skel.boneMap)
        skel.setEntity(entry)
        db.currentDbAccessLog.get(set()).add(skel["key"])
        try:
            yield skel
        except GeneratorExit:
            logger.warning("GeneratorExit. Stop iteration.")
            break



import logging
import time
import typing as t
from types import TracebackType

from viur.core import current

__all__ = ["LanguageContext", "TimeMe"]


class LanguageContext:
    """
    Switch the language of the request just for a specific scope.
    """

    def __init__(self, lang: str):
        self.lang = lang
        self.orig_lang = None

    def __enter__(self) -> t.Self:
        self.orig_lang = current.language.get()
        # TODO: Session?
        current.language.set(self.lang)
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> t.Literal[False]:
        current.language.set(self.orig_lang)
        return False


class TimeMe:
    """
    Measures the execution time of the scope.
    """

    def __init__(self, name: str):
        self.name = name

    def __enter__(self) -> t.Self:
        self.start = time.perf_counter()
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> t.Literal[False]:
        self.end = time.perf_counter()
        logging.debug("%s took %.4fs", self.name, self.end - self.start)
        return False



import logging
import time
import typing as t

from viur.core import bones, db, skeleton

__all__ = [
    "normalize_key",
    "write_in_transaction",
    "increase_counter",
    "set_status",
]

logger = logging.getLogger(__name__)

_KeyType: t.TypeAlias = str | db.Key


def normalize_key(key: _KeyType) -> db.Key:
    if isinstance(key, str):
        return db.Key.from_legacy_urlsafe(key)
    elif isinstance(key, db.Key):
        return key
    raise TypeError(f"Expected key of type str or db.Key, got: {type(key)}")


def write_in_transaction(key: _KeyType, create_missing_entity: bool = True, **values: t.Any) -> db.Entity:
    def txn(_key: db.Key, _values: dict[str, t.Any]) -> db.Entity:
        if (entity := db.Get(_key)) is None:
            if create_missing_entity:
                entity = db.Entity(_key)
            else:
                raise db.NotFoundError(f"Entity {_key} does not exist")
        entity.update(values)
        db.Put(entity)
        return entity

    return db.RunInTransaction(txn, normalize_key(key), values)


def increase_counter(key: _KeyType, name: str, value: float | int = 1, start: float | int = 0) -> int | float:
    """Increase the counter by `value` and return the old value.

    That means the entity holds always the next value.
    """

    def txn(_key: db.Key, _name: str, _value: float | int, _start: float | int) -> float | int:
        if (entity := db.Get(_key)) is None:
            entity = db.Entity(_key)
        if _name not in entity:
            entity[_name] = _start
        old_value = entity[_name]
        entity[_name] += _value
        db.Put(entity)
        return old_value

    return db.RunInTransaction(txn, normalize_key(key), name, value, start)


def set_status(
    key: t.Optional[_KeyType] = None,
    values: dict | t.Callable[[skeleton.SkeletonInstance | db.Entity], None] | None = None,
    precondition: t.Optional[dict | t.Callable[[skeleton.SkeletonInstance | db.Entity], None]] = None,
    create: dict[str, t.Any] | t.Callable[[skeleton.SkeletonInstance | db.Entity], None] | bool = False,
    skel: t.Optional[skeleton.SkeletonInstance] = None,
    update_relations: bool = False,
    retry: int = 1,
) -> skeleton.SkeletonInstance | db.Entity:
    """
    Universal function to set values of an entity within a transaction.
    It is mostly used for status changes, but can also change any value.

    :param key: Entity key to change
    :param values: A dict of key-values to update on the entry, or a callable that is executed within the transaction
    :param precondition: An optional dict of key-values to check on the entry before; can also be a callable.
    :param create: When key does not exist, create it, optionally with values from provided dict, or in a callable.
    :param skel: Use assigned skeleton instead of low-level DB-API
    :param update_relations: Trigger update relations task on success (only in skel-mode, defaults to False)
    :param retry: On ViurDatastoreError, retry for this amount of times.

    If the function does not raise an Exception, all went well.
    It returns either the assigned skel, or the db.Entity on success.
    """
    if key is None and skel is None:
        raise ValueError("No Key is provided")
    elif key is None and skel is not None:
        key = skel["key"]
    if values is None:
        raise ValueError("No values provided")
    # Transactional function
    def transaction() -> skeleton.SkeletonInstance | db.Entity:
        exists = True
        # Use skel or db.Entity
        if skel:
            if not skel.fromDB(key):
                if not create:
                    raise ValueError(f"Entity {key=} not found")

                skel["key"] = key
                exists = False

            obj = skel
        else:
            obj = db.Get(key)

            if obj is None:
                if not create:
                    raise ValueError(f"Entity {key=} not found")

                obj = db.Entity(key)
                exists = False

        # Handle create
        if not exists and create:
            if isinstance(create, dict):
                for bone, value in create.items():
                    obj[bone] = value
            elif callable(create):
                create(obj)

        # Handle precondition
        if isinstance(precondition, dict):
            for bone, value in precondition.items():
                if obj[bone] != value:
                    raise ValueError(f"{bone} contains {obj[bone]!r}, expecting {value!r}")

        elif callable(precondition):
            precondition(obj)

        # Set values
        if isinstance(values, dict):
            for bone, value in values.items():
                # Increment by value?
                if bone[0] == "+":
                    obj[bone[1:]] += value
                # Decrement by value?
                elif bone[0] == "-":
                    obj[bone[1:]] -= value
                else:
                    if skel and (
                        (boneinst := getattr(skel, bone, None))
                        and isinstance(boneinst, bones.RelationalBone)
                    ):
                        assert skel.setBoneValue(bone, value)
                        continue

                    obj[bone] = value

        elif callable(values):
            values(obj)

        else:
            raise ValueError("'values' must either be dict or callable.")

        if skel:
            assert skel.toDB(update_relations=update_relations)
        else:
            db.Put(obj)

        return obj

    # In case of already in transaction, just call the function.
    if db.IsInTransaction:
        return transaction()

    # Otherwise, run the retry loop
    while True:
        try:
            return db.RunInTransaction(transaction)

        except db.ViurDatastoreError as e:
            retry -= 1
            if retry <= 0:
                raise

            logging.debug(f"{e}, retrying {retry} more times")

        time.sleep(1)



import logging
import typing as t  # noqa
from datetime import datetime as dt, timedelta as td, timezone as tz  # noqa

from google.appengine.api.memcache import Client
from google.appengine.ext.testbed import Testbed

from viur.core import conf, utils

__all__ = [
    "MemcacheWrapper",
]

logger = logging.getLogger(__name__)

Seconds: t.TypeAlias = int | float
Args = t.ParamSpec("Args")
Value = t.TypeVar("Value")

if conf.instance.is_dev_server:
    # On the local dev_appserver, we use Google's memcache stub,
    # originally designed for test cases, as a local emulator.
    logger.debug("Using memcache stub")
    testbed = Testbed()
    testbed.activate()
    testbed.init_memcache_stub()

memcache: t.Final = Client()


class MemcacheWrapper(t.Generic[Value, Args]):
    """Wrapper to store computed values in memcache.

    A value will be recalculated after the cachetime has expired
    or the memcached was flushed.
    """

    __slots__ = ("name", "func", "args", "cachetime", "namespace")

    def __init__(
        self,
        func: t.Callable[Args, Value],
        *,
        name: t.Optional[str] = None,
        args: Args.args = tuple(),
        cachetime: td | Seconds = td(hours=1),
        namespace: t.Optional[str] = None,
    ):
        """Initialize a new MemcacheWrapper instance.

        :param func: The function to calculate the value.
        :param name: The name under which the value is to be stored in the memcache.
        :param args: Arguments for the function, must be static
        :param cachetime: Optional expiration time in seconds.
        :param namespace: The namespace in the memcache.
        """
        if name is None:
            name = func.__qualname__
        self.name: str = "/".join([name] + list(map(repr, args)))
        self.func: t.Callable[Args, Value] = func
        self.args: Args.args = args
        self.cachetime: td = utils.parse.timedelta(cachetime)
        if namespace is None:
            namespace = f"cache_{conf.instance.app_version}"
        self.namespace: str = namespace

    def get(self) -> Value:
        """Get the value from memcache

        Trigger the recalculation if necessary.
        """
        res = memcache.get(self.name, namespace=self.namespace)
        # logger.debug("res for %r: %r", self, res)
        if res is None:
            res = self.set()
        return res

    def set(self) -> Value:
        """Set the value (force a recalculation) in the memcache"""
        res = self.func(*self.args)
        memcache.set(self.name, res, self.cachetime.total_seconds(), namespace=self.namespace)
        return res

    def clear(self) -> t.Any:
        """Drop the stored value in memcache"""
        return memcache.delete(self.name, namespace=self.namespace)

    def __repr__(self) -> str:
        return "<%s.%s object, name=%r, namespace=%r, func=%r, args=%r, cachetime=%r>" % (
            self.__class__.__module__, self.__class__.__name__,
            self.name, self.namespace, self.func, self.args, self.cachetime,
        )



import functools
import operator

import typing as t

__all__ = [
    "NOT_SET",
    "vars_full",
    "freeze_dict",
]


class NotSet:
    def __repr__(self) -> str:
        return "<NOT_SET>"

    def __bool__(self) -> bool:
        return False


NOT_SET: t.Final[NotSet] = NotSet()


def vars_full(
    obj: t.Any,
    *,
    include_properties: bool = True,
) -> dict[str, t.Any]:
    """
    A better version of the builtin `vars()`.

    This method is designed for debugging.
    It returns the members of an instance, regardless of whether it uses
    `__dict__` or `__slots__`.
    It also adds the values of properties.
    """
    res = {}
    if hasattr(obj, "__slots__"):
        for slot in obj.__slots__:
            res[slot] = getattr(obj, slot, NOT_SET)
    elif hasattr(obj, "__dict__"):
        res |= vars(obj)
    if not include_properties:
        return res
    for attr, value in vars(type(obj)).items():
        if isinstance(value, (property, functools.cached_property)):
            res[attr] = getattr(obj, attr, NOT_SET)
    return res


def freeze_dict(value: dict[str, t.Any]) -> list:
    """Sort a dict recursively by keys and return as list"""
    return sorted(
        [
            (pair[0], freeze_dict(pair[1])) if isinstance(pair[1], dict) else pair
            for pair in value.items()
        ],
        key=operator.itemgetter(0),
    )



from viur.core import conf, current, db

__all__ = ["contains_substring", "user_has_access"]


def contains_substring(haystack: str, *needles: str) -> bool:
    """Check if *haystack* contains one of the *needles* as substring."""
    return any(needle in haystack for needle in needles)


def user_has_access(*roles: str, user: db.Key | None = None) -> bool:
    """Check if the user has one of the given roles."""
    if user is None:  # this None acts as sentinel
        user = current.user.get()
        if user is None:
            return False  # guests have no access
    elif isinstance(user, db.Key):
        _key = user
        user = conf.main_app.vi.user.viewSkel()
        user.fromDB(_key)
    else:
        raise ValueError(f"Invalid value for user: {user!r}")

    return any(role in user["access"] for role in roles)



__all__ = ["round_decimal", "format_currency"]


def round_decimal(value: float, frac_digits: int = 2) -> float:
    """Round decimal correctly, even as floating-point"""
    return round(value * (10 ** frac_digits), 0) / (10 ** frac_digits)


def format_currency(value: float, frac_digits: int = 2) -> str:
    before, after = "{:,.2f}".format(round_decimal(value, frac_digits)).split(".")
    before = before.replace(",", ".")
    return f"{before},{after}"
